from . import Rbeast as cb

def print(o, index = 0):
    """
   USAGE: printbeast(o, index)

   o :  the time series analysis output from  beast123; o
   should contain results for mulltiple time series

   index: if o contains results for more than 1 time
   series, index specifies for which time series the result is printed.
   If o is the result for a 3D stacked cube, index will be a vector of 2
   integer to specify the row and col of the desired pixel. If o contains
   only one time series, index will be ignored

  Contact info : To report bug or get help, do not hesitate to contact Kaiguang Zhao
   at <strong>zhao.1423@osu.edu</strong>.
    """
    if hasattr(o, 'marg_lik'):
        if isinstance(index, int):
           index=index+1
        else:
           index = [x+1 for x in index]   
        cb.Rbeast('print',o,index);

def obj_repr(o):    
    N    = len( o.__dict__.keys())
    s    = "Object of %d field(s):\n\n" % (N)
    s    = my_own_repr(o, s ,0+1)
    return s
   
import numpy as np   
def my_own_repr(o, s, left):

    dict           = list( o.__dict__.keys())
    maxFldLen      = max([len(x) for x in dict])
    N              = len(dict)
    for i in range( N ):
        key = dict[i]
        s = s +'%*s' % (left,'')+ "%-*.*s : " % (maxFldLen,maxFldLen, key)
        e = o.__dict__[key]
        if isinstance(e, np.ndarray):
            etype = e.dtype
            if e.size == 1:
                if etype == np.int16 or etype== np.int32 or etype== np.int64:
                   s = s + "%d \n" % (e[0])
                elif etype == np.float32 or etype== np.float64:
                   s = s + "%g \n" % (e[0])
                else:
                   s = s + "%g \n" % (e[0])
                continue

            if  etype == np.int16:
                typestr ='int16'
            elif etype == np.int32:
                typestr = 'int32'
            elif etype == np.int64:
                typestr = 'int64'
            elif etype == np.float32:
                typestr = 'float32'
            elif etype== np.float64:
                typestr = 'float64'
            else:
                typestr = 'others'

            dims = e.shape
            tmp = "[%d" % ( dims[0])
            for j in range(len(dims)-1):
                tmp = tmp + "x%d" % (dims[j+1])
            s = s +tmp +" " + typestr + "] \n"
        #if isinstance(e, BeastOutputClass):
        if isinstance(e, type(o)):        
            s = s+ "[ 1 object with %d fields] \n" % (len( e.__dict__.keys()))
            s = my_own_repr(e,s, left+ maxFldLen+3)
        if isinstance(e, str):
            s = s+ "'"+e+ "'\n"
    return s


#cbeast.pyobject.__getitem__ = extract




from . import Rbeast as cb
from .cvt_to_numpy import force_convert_to_numpy  ### Y = force_convert_to_numpy(Y)

def beast123( Y,                  # Y can be 1D (one time series), 2D (many ts), or 3D (stacked images)
              metadata = None,    # an object with many fields to describe Y 
              prior    = None,    # an object with many fields to specify the bayesian model
              mcmc     = None,    # an object to specify the MCMC simulation parameters 
              extra    = None,    # an object to specify the outputs 
              method   = 'bayes', # 'bayes','bic','aic','aicc','hic', 'bic0.25','bic0.5'
              **kwargs ):
      """
   
   
######################################################################################################
Bayesian changepoint detection and time series decomposition for regular or irregular time series data
    
The fitted model is:
     Y=trend+error             if data has no periodic/seasonal variation (i.e., season='none')
     Y=trend+seasonal+error    if data has periodic/seasonal variation 
     Y=trend+outlier +error    if data is trend-only (no seasonal variation) but with potential outliers
     Y=trend+seasonal+outlier+error if data has periodic/seasonal variation and also has outliers
where trend is a piecewise linear or polynomial function with an unknown number of trend changepoints to 
be inferred; seasonal is a piecewise periodic function with an unknown number of seasonal changepoints to 
be inferred; and the outlier component refers to potential pikes or dips at isolated data points and is 
included only if metadata.hasOutlierCmpnt=True (in beast123) or hasOutlier=True (in beast or beast_irreg)
######################################################################################################

--------------------------------------------------------------------------------------------------
*Note*:
--------------------------------------------------------------------------------------------------
(1) beast123() is an all-inclusive function that duplicates the functionalities of beast() 
    and beast_irreg(). beast123() can handle a single, multiple, or 3D stacked time series, being
    either regular or irregular.  beast123() internally supports parallel computing if the input is mutliple time
    series or stacked time series images. It allows for customization through four object argumentsa: metadata,
    prior, mcmc, and extra.

(2) Currently, the BEAST algorithm is implemented to handle regular time series purely for computational 
   advantages. The function 'beast_irreg' does accept irregular time series but internally it still 
   aggregates them into regular ones prior to applying the BEAST model. For the aggregation, both the "time"
   and "deltat" args are needed to specify indvidial times of data points and the regular time interval desired. 

--------------------------------------------------------------------------------------------------
*Quick Examples*:
--------------------------------------------------------------------------------------------------
import Rbeast as rb
    
nile,yr =  rb.load_example('nile')                   # annual flow of the Nile river
rb.beast123( nile, metadata=rb.args(season='none') ) # season='none' explicitly provided via metadata
rb.beast123( nile, metadata=rb.args(season='none',start=1871,deltat=1.0) ) # metadata passing more info
   
ohio = rb.load_example('ohio')                # a satelite ndvi time series at an Ohio site
meta          = rb.args()  # Just an empty object to stuff more fields
meta.time    = ohio.rdate  # times of individual data points 
meta.deltat  = 1/12        # desired regular time interval when aggragating the irregular time series  
meta.period  = 1.0         # the period of the time series (i.e., with a seasonal component)
extra        = rb.args( quiet=True, computeTrendSlope=True ) # Directly add fields to specify parameters
o    = rb.beast123(ohio.ndvi, metadata=meta, extra=extra)
rb.plot(o) 
rb.plot(o, vars=['s', 't','tcp','tslp','slpSgn'])
    
--------------------------------------------------------------------------------------------------    
*Input arguments*:
--------------------------------------------------------------------------------------------------
Y:  an iregularly-spaced time series or 1D signal; it should be a numeric vector. For reggular 
    time series, use 'beast' or 'beast123' instead. For multiple time series  or stacked time series 
    images such as satellite data, use 'beast123'.
 
... : the remaining arguments are many paired keywords and values to specifiy time information 
    or parameters for the beast algorithm. Check the R version of BEAST for detailed explanations
    (https://cran.r-project.org/web/packages/Rbeast/Rbeast.pdf). Below is a brief description.
 
--------------------------------------------------------------------------------------------------    
*Possible Keywords*:
--------------------------------------------------------------------------------------------------   
 
|> metadata <| a struct variable specifying metadata for the input Y 

     metadata.season           : a string specifier about the seasonal component. Possible values are
	                             (1) 'none'     - trend-only data with no seasonlity or periodic variation,
                                 (2) 'harmonic' - harmonic model for the seasonal component,     
                                 (3) 'dummy'    - dummy model for the seasonal component     
                                 (4) 'svd':     - (experimental feature) the same as 'harmonic' except that the 
                                  periodic/seasonal component is modeled as a linear combination of function 
                                  bases derived from a Single-value decomposition. The SVD-based basis functions 
                                  are more parsimonious than the harmonic sin/cos bases in parameterizing the 
                                  seasonal variations; therefore, more subtle changepoints are likely to be detected.
     metadata.period           : the period of the seasonal/periodic component
     metadata.isRegularOrdered : Deprecated. Now use 'metadata.startTime' and ' metadata.deltaTime' to 
                                 specify times for evenly-spaced data, and use ' metadata.time' to
                                 provide times of individual data points for irregular (or regular) data	 
     metadata.startTime        : the start time of regular input
     metadata.deltaTime        : the time interval between consecutive data points. For regular time seires, it
                                 determins the times of all the data points; for irregular time series, deltaTime
                                 specifies the interval at which the irregular input is aggregated/resampled
     metadata.time             : for irregular inputs (or regular unordered data), specify idividual times of 
                                 data points. Both string and numeric vectors are allowed; for more details,
                                 check rdrr.io/cran/Rbeast/man/beast123.html.
     metadata.whichDimIsTime   : which dim of the 2D or 3D input refer to time
     metadata.missingValue     : a value indicating bad/missing data
     metadata.maxMissingRate   : the max missingness rate beyond which the ts will be ignored
     metadata.deseasonalize    : if true, the input ts is first de-seasonalize by removing a global seasonal component,
                                 prior to applying BEAST
     metadata.detrend          : if true, the input ts is first de-trended by removing a global trend component, 
                                 prior to applying BEAST
     metadata.hasOutlierCmpnt  : boolean; if true, fit a model with an outlier component.
                                 Y = trend + outlier + error if season='none' or
                                 Y = trend+season+outlier+error will be fitted.
								 
 
|> prior <|  prior hyperparameter for the BEAST model 
  
     prior.seasonMinOrder   : min harmonic order considered for the seasonal compnt
     prior.seasonMaxOrder   : max harmonic order considered for the seasonal compnt
     prior.seasonMinKnotNum : min number of seasonal changepoints allowed
     prior.seasonMaxKnotNum : max number of seasonal changepoints allowed
     prior.seasonMinSepDist : the min segment length of a seaosnal segment (i.e, the min seperation
                              distance between neighorboring seasonal changepoints)
     prior.seasonLeftMargin : an integer;  the number of leftmost data points excluded for 
	                          seasonal changepoint detection. That is,  no seasonal changepoints are allowed in 
	                          the starting window/segment of length seasonLeftMargin. seasonLeftMargin must
	                          be an unitless integer–the number of time intervals/data points so that the
	                          time window in the original unit is seasonLeftMargin*deltat. If missing,
	                          seasonLeftMargin defaults to the minimum segment length 'seasonMinSepDist'
     prior.seasonRightMargin: an integer;  the number of rightmost data points excluded for 
	                          seasonal changepoint detection. That is,  no changepoints are allowed in 
	                          the ending window/segment of length seasonRightMargin. seasonRightMargin must
	                          be an unitless integer–the number of time intervals/data points so that the
	                          time window in the original unit is seasonRightMargin*deltat. If missing,
	                          seasonRightMargin defaults to the minimum segment length 'seasonMinSepDist'
     prior.trendMinOrder	: min polyonimal order considered for the trend compnt
     prior.trendMaxOrder	: max polyonimal order considered for the trend compnt
     prior.trendMinKnotNum  : min number of trend changepints allowed
     prior.trendMaxKnotNum  : max number of trend changepints allowed
     prior.trendMinSepDist  : the min segment length of a trend segment(i.e, the min seperation distance
                             between any two neighorboring trend changepoints)
     prior.trendLeftMargin  : an integer;  the number of leftmost data points excluded for 
	                         trend changepoint detection. That is,  no trend changepoints are allowed in 
	                         the starting window/segment of length trendLeftMargin. trendLeftMargin must
	                         be an unitless integer–the number of time intervals/data points so that the
	                         time window in the original unit is trendLeftMargin*deltat. If missing,
	                         trendLeftMargin defaults to the minimum segment length 'trendMinSepDist'
     prior.trendRightMargin : an integer;  the number of rightmost data points excluded for 
	                         trend changepoint detection. That is,  no trend changepoints are allowed in 
	                         the ending window/segment of length trendRightMargin. trendRightMargin must
	                         be an unitless integer–the number of time intervals/data points so that the
	                         time window in the original unit is trendRightMargin*deltat. If missing,
	                         trendRightMargin defaults to the minimum segment length 'trendMinSepDist'
     prior.outlierMinKnotNum: an integer; needed only if meta.hasOutlierCmpnt=True to specify the minimum 
                             number of outliers (i.e., outlier-type changepoints) allowed in the time series	
     prior.outlierMaxKnotNum: an integer; needed only if meta.hasOutlierCmpnt=True to specify the maximum 
                             number of outliers (i.e., outlier-type changepoints) allowed in the time series	
     prior.precValue        : numeric (>0); the hyperparameter of the precision prior; the default value is 1.5. 
                             precValue is useful only when precPriorType='constant', as further explained below
     prior.precPriorType    :  a string taking one of 'constant', 'uniform',  'componentwise' (default), and 'orderwise'.
                             (1) 'constant':  the precision parameter used to parameterize the model coefficients is fixed to
                             a const specified by precValue. In other words, precValue is a user-defined hyperparameter 
                             and the fitting result may be sensitive to the chosen values of precValue.
                             (2) 'uniform':  the precision parameter used to parameterize the model coefficients is a random variable;
                             its initial value is specified by precValue. In other words, precValue will be inferred by the MCMC,
                             so the fitting result will be insensitive to the chose inital value of precValue.
                             (3) 'componentwise': multiple precision parameters are used to parameterize the model coefficients for
                             individual components (e.g., one for season and another for trend); their initial values is specified 
                             by precValue. In other words, precValue will be inferred by the MCMC, so the fitting result will be 
                             insensitive to the choice in precValue.
                             (4) 'orderwise'}: multiple precision parameters are used to parameterize the model coefficients not just for 
                             individual components but also for individual orders of each component; their initial values is specified 
                             by precValue. In other words, precValue will be inferred by the MCMC, so the fitting result will be 
                             insensitive to the choice in precValue. 							 

|> mcmc <|    parameters to set up the MCMC sampler
 
     mcmc.seed                      : the seed for the random number generator            
     mcmc.samples                   : the number of samples to be collected
     mcmc.thinningFactor            : a thinning factor; take every 'thinningFactor'-th sample from the chain
     mcmc.burnin                    : number of initial samples discarded
     mcmc.chainNumber               : number of chains
     mcmc.maxMoveStepSize           : a moving step
     mcmc.trendResamplingOrderProb  : probability to re-sample trend order
     mcmc.seasonResamplingOrderProb : probability to re-sample seasonal order
     mcmc.credIntervalAlphaLevel    : signifiance level for computing  credible interval
 
|> extra <|    parameters to control computaiton or outputs

     extra.dumpInputData       : if true, dump the input time series (o.data)
     extra.whichOutputDimIsTime: which dim of the output array refer to time
     extra.computeCredible     : compute credible intervals (unless needed for assessing the curve fitting, 
	                             it is better to disable it for changepoint analysis because computing CI is
	                             expensive.)(o.trend.CI and o.season.CI)
     extra.fastCIComputation   : if true, a fast version of algorithm used to compute CI. But it is still
	                             expensive. Set extra.computeCredible=false if CI is of litte  interest
     extra.computeSeasonOrder  : if true, estimate harmonic orders needed to adequately fit the seasonal
                                 component (o.season.order)
     extra.computeTrendOrder   : if true, estimate polynomial orders needed to adequately fit the trend 
                                 component(o.trend.order)
     extra.computeSeasonChngpt  : if true, dump the scp detected (o.season.cp, o.season.cpCI, and cpAbruptChange)
     extra.computeTrendChngpt   : if true, dump the tcp detected (o.trend.cp, o.trend.cpCI, and cpAbruptChange)
     extra.computeSeasonAmp     : if true, estimate the seasonal ammplitue (o.seasin.amp)
     extra.computeTrendSlope    : if true, estimate time-varying slope and the associated probabilities of slopes 
                                  being positive or negative (trend.slp, trend.slpSgnPosPr, trend.slpSgnZeroPr)
     extra.tallyPosNegSeasonJump: boolean
     extra.tallyPosNegTrendJump : boolean
     extra.tallyIncDecTrendJump : boolean
     extra.printParameter        : boolean; if true, print the BEAST parameters at the start	 
     extra.printProgress        : boolean; if true, print a progress bar
     extra.printWarning         : boolean; if true, print warning messages
     extra.quiet                : boolean; if true, print nothing
     extra.consoleWidth         : the console/terminal width
     extra.numThreadsPerCPU     : IMPORTANT for remote sensing/earth science applicaitons: specify the numbers 
                                  of concurrent threads per cpu core
     extra.numParThreads        : specify the number of total concurrent threads

|> method <|  a string specifying which method to formulate model posterior probability. Possible values are 

    (1) 'bayes': the full Bayesian formulation (this is the default)  
    (2)'bic':  approximated posterior probability using the Bayesian information criterion (bic)
    (3)'aic':  approximated posterior probability using the Akaike information criterion (aic)
    (4)'aicc': approximated posterior probability using the corrected Akaike information criterion (aicc)
    (5)'hic':  approximated posterior probability using the Hannan–Quinn information criterion  (hic)
    (6)'bic0.25':  approximation using the Bayesian information criterion adopted from Kim et al. (2016) <doi: 
              10.1016/j.jspi.2015.09.008>; bic0.25=n*ln(SSE)+0.25k*ln(n) with less complexity penelaty than the standard BIC.
    (7)'bic0.50': the same as above except that the penalty factor is 0.50.
    (8)'bic1.5':  the same as above except that the penalty factor is 1.5.
    9)'bic2':    the same as above except that the penalty factor is 2.0.
         
--------------------------------------------------------------------------------------------------		 
*Extra Note*:
--------------------------------------------------------------------------------------------------
beast, beast_irreg, and beast123 calls the same dll extension library written in C/C++ (i.e., Rbeast.pyd);
they are just wrappers to the core BEAST algorithm. In particular, BEAST currently handles only regular time
series; irregular inputs will be first aggregated  before applying BEAST.

--------------------------------------------------------------------------------------------------
*Result/Output*: The output is a struct variable; example of the fields include
--------------------------------------------------------------------------------------------------
    
        marg_lik: marginal likilood; the larger, the better
        sig2    : variance  of error
        trend   : the trend component; a struct variable (say, T)
        season  : the season componet; a stuct variable  (say,S)
        The subfields of trend or season:
        .ncpPr        : the prob distribution for number of changepoints
        .ncp          : mean number of changepoints in trend or seasonality
        .ncp_meidan   : median number of changepoints
        .ncp_mode     : mode from ncpPr
        .ncp_pct90    : 90# percentile from ncpPr
        .cpOccPr      : changepoint occurrance probability over time
        .cp           : list of all possible changepoints (many are not sigficant)
        .cpPr         : occurrence probability of the changepoints in cp
        .cpAbruptChange: the sudden changes in trend or seasonlity at cp
        .cpCI         : confidence interval of the cps
        .Y            : the fitted trend or seasonality 
        .SD           : standard deviation of the fitted Y
        .CI           : Credible interval of the fittted Y
        .order   : the mean harmonic or polynomial orders estimated to fit the seasonal and trend     
        trend.slp     : slope of the trend 
        trend.slpSD   : standard dev of the estimated slope
        trend.slpSgnPosPr: time-varying probability of the slope being postive
        trend.slpSgnZeroPr: time-varying probability of the slope being 0
        season.amp     : amplitue of the estiamted seasonality overtime
        season.ampSD   : standard ev of the estiamated amplitude

-------------------------------------------------------------------------------------------------- 
More help:  
--------------------------------------------------------------------------------------------------
       This terse help doc sucks (I know); so far, the best details are still the
       R help doc, available at https://cran.r-project.org/web/packages/Rbeast/Rbeast.pdf.
       Python doesn't allow a '.' in variable names, so Python's equivalent to R's 
       beast(Y,start=1987,tcp.minmax=c(0,5)) is beast(Y, start=1987,  tcp_minmax=[0, 5]).
       
--------------------------------------------------------------------------------------------------
Examples:
--------------------------------------------------------------------------------------------------
import Rbeast as rb

Nile, year = rb.load_example('Nile') # Nile river annual streamflow: trend-only data
metadata   = rb.args()               # an empty object to suff more attribute fields
metadata.season    = 'none'          # trend-only
metadata.startTime = 1871;
metadata.deltaTime = 1;
       
o = rb.beast123(Nile,metadata)       # Default values will be used if parameters are missing
rb.print(o)
rb.plot(o)
 
ohio = rb.load_example('ohio')     # irregular Landsat NDVI time series
metadata = lambda:None;            # equivalent to 'rb.args()', creating an empty dummy object
metadata.season     = 'harmonic'   # seasonality is present
metadata.time       = ohio.time;
metadata.deltaTime  = 1/12;        # aggregate into a monthly ts
metadata.period     = 1.0;         # period= 1/12 x 12 (freq)

o = rb.beast123(ohio.ndvi,metadata)# Default values will be used if parameters are missing
rb.print(o)
rb.plot(o, ncpStat='median')       
 
ohio.datestr1   # strings of times for ohio.ndvi
metadata              = lambda:None # equivalent to 'rb.args()', creating an empty dummy objec
metadata.time         = rb.args( dateStr=ohio.datestr1, strFmt='????yyyy?mm?dd' )
metadata.deltaTime    = 1/12;    # aggregate into a monthly ts
metadata.period       = 1.0;     # period= 1/12 x 12 (freq)
# Default values will be used if parameters are missing
o =  rb.beast123(ohio.ndvi,metadata,[],[], rb.args(dumpInputData=True)) 
rb.plot(o, ncpStat='median')   
 
ohio.datestr2      # strings of times for ohio.ndvi
metadata              = lambda:None
metadata.time         = lambda:None
metadata.time.dateStr = ohio.datestr2;
metadata.time.strFmt  = 'LC8-yyyydoyxdvi';
metadata.deltaTime    = 1/12;   # aggregate into a monthly ts
metadata.period       = 1.0;    # period= 1/12 x 12 (freq)
o = rb.beast123(ohio.ndvi,metadata) 
rb.plot(o, ncpStat='median')  

#See https://rdrr.io/cran/Rbeast/man/beast123.html for more details
#about the accepted formats of date strings
 
simdata = rb.load_example('simdata')      # a 774x3 matrix: 774 is the time dimesnion      
# simdata is a toy example of 3 time series: I decides to be lazy here bcz the 3
# time series are essentially the same. We just assume they are
# different for illustring the use of beast123 to handle 2d matrix      
 
metadata = lambda:None  
metadata.whichDimIsTime = 1   # 774 is the ts length 
o = rb.beast123(simdata ,metadata) 
rb.print(o,  1)            #print the result for the first ts
rb.plot(o, index=1)        #plot the result for the first ts
rb.print(o,  2)            #print the result for the 2nd ts
rb.plot(o, index=2)        #plot the result for the 2nd ts      
 
import numpy as np
simData3x774 = np.transpose( simdata ) # a 3x774 matrix: 774 is the time dimesnion      
metadata = lambda:None
metadata.whichDimIsTime = 2 # 774 is the ts length 
o = rb.beast123(simData3x774 ,metadata) 
rb.print(o,  2) #print the result for the 2nd ts
rb.plot(o, index=2)     #plot the result for the 2nd ts   
 
ndvi3d, fnames, fyear = rb.load_example( 'imagestack' ) 
# A toy example of stacked time series images: unevely-spaced in time
ndvi3d    # a 12x9x1066 3D cube, and 1066 is the time series length
fnames    # a vector of image filenames as date strings
fyear     # a vector of fractional year 
metadata = lambda:None
metadata.time         = lambda:None ;
metadata.time.dateStr = fnames
metadata.time.strFmt  = 'LT05_018032_20110726.yyyy-mm-dd';
metadata.deltaTime    = 1/12; # aggregated at a monthly interval
metadata.period= 1.0;  # the period is 1.0 (year)
extra   =  rb.args();
extra.dumpInputData    = True # get a copy of the aggregated input
extra.numThreadsPerCPU = 2;  # 2 threads per CPU core
o = rb.beast123(ndvi3d,metadata,[],[], extra) 
rb.print(o,[2,4])      #print the result at row 3 and col 5    
rb.plot(o,index=[2,4]) #plot the result at row 3 and col 5    

metadata = lambda:None
metadata.time         = fyear;
metadata.deltaTime    = 1/12; # aggregated at a monthly interval
metadata.period= 1.0;  # the period is 1.0 (year)
extra =  lambda:None
extra.dumpInputData    = true # get a copy of the aggregated input
extra.numThreadsPerCPU = 2;   # 2 threads per CPU core
o = rb.beast123(ndvi3d,metadata,[],[], extra) 
rb.print(o,[2,4])      #print the result at row  3 and col 5    
rb.plot(o,index=[2,4]) #plot the result at row  3 and col 5     
 
Contact info: To report bug or get help, do not hesitate to contact Kaiguang Zhao
at zhao.1423@osu.edu.
      """
      Y = force_convert_to_numpy(Y)
      
      if hasattr(metadata, 'time'):
            time = metadata.time
            if hasattr(time, "year"):
                time.year = force_convert_to_numpy(time.year)
                if hasattr(time, "month"):
                    time.month = force_convert_to_numpy(time.month)
                if hasattr(time, "day"):
                    time.day   = force_convert_to_numpy(time.day)   
                if hasattr(time, "doy"):
                    time.doy   = force_convert_to_numpy(time.doy)   
            elif hasattr(time,'datestr'):
                time.datestr   = force_convert_to_numpy(time.datestr)  
            elif hasattr(time,'dateStr'):
                time.dateStr   = force_convert_to_numpy(time.dateStr) 
                pass
            else: 
                time = force_convert_to_numpy(time)             #then, we assume time is a numeric vector
                
            metadata.time = time
            
      o = cb.Rbeast('beast_' + method,Y, metadata, prior, mcmc, extra)
      return (o)





from . import Rbeast as cb

def extract(o, index):

# USAGE: <strong>extractbeast(o, index) </strong>
#
#   <strong>o</strong>:  the time series analysis output from  beast123; o
#   should contain results for mulltiple time series
#
#   <strong>index </strong>: if o contains results for more than 1 time
#   series, index specifies for which time series the result is extracted.
#   If o is the result for a 3D stacked cube, index will be a vector of 2
#   integer to specify the row and col of the desired pixel. If o contains
#   only one time series, index will be ignored

    if len(o.marg_lik) == 1:
        return o
    else:
        if isinstance(index, int):
             index=index+1
        else:
             index = [x+1 for x in index]  
        # cb.Rbeast uses 1-based indices but Python uses zero-based indices
        return cb.Rbeast('tsextract', o, index);



from   .extractbeast     import extract as extractbeast
from numpy               import ndarray, min, max,ndarray,flip, isnan, round, sum, arange
from numpy               import concatenate as c, array as np_array, where as np_where, sqrt as np_sqrt

import importlib.util
import sys
################################################################
# Check if the matplotlib is installed or not: If yes, load it
################################################################

def check_module(name):
# https://stackoverflow.com/questions/1051254/check-if-python-package-is-installed
  if name in sys.modules:
    return True
  else:
    spec = importlib.util.find_spec(name)
    if spec is not None:
      return True
     #If you choose to perform the actual import ...
     #module = importlib.util.module_from_spec(spec)
     #sys.modules[name] = module
     #spec.loader.exec_module(module)
     #print(f"{name!r} has been imported")      
    else:
      return False

if check_module('matplotlib'):
    import matplotlib.pyplot as plt
    from    matplotlib       import __version__  as version


  
# https://stackoverflow.com/questions/990422/how-to-get-a-reference-to-current-modules-attributes-in-python
#As previously mentioned, globals gives you a dictionary as opposed to dir() which gives you a list of the names 
# defined in the module. The way I # typically see this done is like this:
# import sys
# dir(sys.modules[__name__])  

def is_matlot_loaded():   
   return 'plt' in globals()



import warnings

VARS    = ['st', 's', 't', 'scp', 'tcp', 'sorder', 'torder', 'error', 'o', 'ocp', 'samp', 'tslp', 'slpsgn']
YLABELS = {'st': 'Y', 's': 'season', 't': 'trend', 'o': 'outlier', 'scp': 'Pr(scp)', 'tcp': 'Pr(tcp)', 'ocp': 'Pr(ocp)',
           'sorder': 'sOrder', 'torder': 'tOrder', 'samp': 'amplitude', 'tslp': 'tslp', 'slpsgn': 'slpsgn',
           'error': 'error'}
COLORS  = {'st': [0.1, 0.1, 0.1], 's': 'r', 't': 'g', 'o': 'b', 'scp': 'r', 'tcp': 'g', 'ocp': 'b', 'sorder': 'r',
        'torder': 'g', 'samp': 'r', 'tslp': 'g', 'slpsgn': 'k', 'error': [0.4, 0.4, 0.4]}
HEIGHTS = {'st': 0.8, 's': 0.8, 't': 0.8, 'o': 0.8, 'scp': .5, 'tcp': .5, 'ocp': .5, 'sorder': .5, 'torder': .5,
           'samp': .4, 'tslp': .4, 'slpsgn': .4, 'error': .4}

length  = len
isfield = hasattr


def getfield(obj, fld):
    if isinstance(obj, dict):
        return obj[fld]
    if hasattr(obj, '__dict__'):
        return obj.__dict__[fld]
    return None


def isempty(obj):
    if obj is None:
        return True
    if isinstance(obj, ndarray):
        return len(obj) == 0
    if hasattr(obj, '__dict__'):
        return len(obj.__dict__) == 0
    return len(obj) == 0


def plot(o, index=0,\
         vars       = ["st", "s", "scp", "sorder", "t", "tcp", "torder", "slpsgn", "o", "ocp", "error"], \
         ncpStat    = 'median', \
         relheights = [],\
         fig        = [], \
         title      = 'BEAST decompositon and changepoint detection',\
         ylabels    = [],\
         xlabel     = 'time'\
         ):

    """    
--------------------------------------------------------------------------------------------------           
*Input arguments*:
--------------------------------------------------------------------------------------------------
   
o:    the time series analysis output from  beast123; o
      should contain results for mulltiple time series

index: if o contains results for more than 1 time series, index specifies  for which 
       time series the result is printed.  If o is the result for a 3D stacked cube, 
       index will be a vector of 2  integer to specify the row and col of the desired pixel.
       If o contains  only one time series, index will be ignored
 
vars:  a vector of strings indicating the elements or variables 
       of o to plot. Possible vars strings include 'st' (season plus trend), 's' (season component),
       't' (trend component), 'o' (outliers), 'scp', 'tcp', 'ocp'  (occurrence probability
       of seasonal/trend/outlier changepoint), 'sorder' (seasonal harmonic order), 
       'torder' (trend polynomial order), 'samp' (amplitude of seasonality), 'tslp' (slope of trend),
       'slpsgn' (probabilities of the slope being positive, zero, and negative) and 'error' (remainder).
   
ncpStat: a string to specify which statistic to use for the number of ChangePoint (ncp). 
         Five values are possible: 'mean', 'mode', 'median', 'pct90', and 'max'; the default is 'median'. 
         Individual models sampled by BEAST has a varying dimension (e.g., number of changepoints or knots). 
         For example, if mcmc$samples=10, the numbers of changepoints for the 10 sampled models are assumed
         to be c(0, 2, 4, 1, 1, 2, 7, 6, 6, 1). 
		 (1) the mean ncp will be 3.1 (rounded to 3), 
		 (2) the median is 2.5 (2)
		 (3) the mode is 1, 
		 (4) the maximum is 7.
         The 'max' option will plot all the changepoints recorded in out.trend.cp, out.season.cp, or o.outlier.cp; 
         many of these changepoints are bound to be false positives, so do not blindly treat all of them as 
         actual changepoints.

-------------------------------------------------------------------------------------------------- 
Examples:
--------------------------------------------------------------------------------------------------
    # in the slpsgn plot, the upper red envelope refers to probability of trend slope being positive,
    # the middle green to probability of trend slope being zero, the blue to probability of slope
    # of being negative

    import Rbeast as rb
    Nile, time = rb.load_example('nile')
    o          = rb.beast(Nile, season='none')
    rb.plot( o , vars = ["st", "t", "tcp" ,"slpsgn"], ncpStat='mean' )
  
    rb.plot(o, index = 3)
    rb.plot(o, index = [3,5])
 
    Contact info: To report bug or get help, do not hesitate to contact Kaiguang Zhao
    at zhao.1423@osu.edu.
    """         

#    index = 0;
#    vars = ["st", "s", "scp", "sorder", "t", "tcp", "torder", "slpsgn", "o", "ocp", "error"]
#    relheights = []
#    ylabels = []
#    xlabel = 'time'

    class Obj:
        pass

    def error(msg):
        raise ValueError(msg)

    def warning(msg):
        warnings.warn(msg)

    if not is_matlot_loaded():
        error("matplotlib is needed for plotting but not installed! Install it before using this function!")
        return None
    #indexDefautValue    = 0;
    #varsDefautValue     = ["st", "s", "scp", "sorder", "t", "tcp", "torder", "slpsgn", "o", "ocp", "error"]
    #ncpStatDefaultValue = 'median'
    #vars    = varsDefautValue
    #ncpStat = ncpStatDefaultValue
    ncpStat = ncpStat.lower()
    vars    = [x.lower() for x in vars]
    varsold = vars.copy()
    vars    = [x for x in varsold if x in VARS]

    if  len(relheights) < len(varsold):
        heights = [HEIGHTS[key] for key in vars]
    else:
        heights = relheights[0:len(varsold)]
        heights = [heights[i] for i in range(len(varsold)) if varsold[i] in vars]

    if  len(ylabels) < len(varsold):
        ylab =  [YLABELS[key] for key in vars]
    else:
        ylab = ylabels[0:len(varsold)]
        ylab = [ylab[i] for i in range(len(varsold)) if varsold[i] in vars]
    # ylab    = {key:YLABLES[key] for key in vars}

    col     = [COLORS[key] for key in vars]


    if  length(o.marg_lik)> 1 :
        # more than time series is present
        o=extractbeast(o,index);        

    if isnan(o.marg_lik):
        H = [];
        warning("The result of the selected time series selected is invalid.");
        return None

    # %%
    hasData   = isfield(o, "data")   and  not  isempty(getfield(o, 'data'));
    hasSeason = isfield(o, 'season') and  not  isempty(o.season);
    if hasSeason:
        hasSOrder = isfield(o.season, 'order') and not  isempty(o.season.order);
        hasAmp    = isfield(o.season, 'amp')   and not isempty(o.season.amp);
    else:
        hasSOrder = False;
        hasAmp    = False;

    hasOutlier = isfield(o, 'outlier') and not isempty(o.outlier);
    hasTOrder  = isfield(o.trend, 'order') and not isempty(o.trend.order);
    hasSlp     = isfield(o.trend, 'slp') and not  isempty(o.trend.slp);

    has = Obj();
    has.hasAmp = hasAmp;
    has.hasSlp = hasSlp;
    has.hasSeason = hasSeason;
    has.hasSOrder = hasSOrder;
    has.hasTOrder = hasTOrder;
    has.hasOutlier = hasOutlier;
    has.hasData = hasData;

    # %%
    idx = [True for x in vars];
    if (not hasAmp):
        idx = [idx[i] if vars[i] != 'samp' else False for i in range(len(idx))]
    if (not hasSlp):
        idx = [idx[i] if vars[i] != 'tslp' and vars[i] !='slpsgn' else False for i in range(len(idx))]
    if (not hasSeason):
        idx = [idx[i] if vars[i] != 'st'     and vars[i] != 's' else False for i in range(len(idx))]
        idx = [idx[i] if vars[i] != 'sorder' and vars[i] != 'scp' else False for i in range(len(idx))]
    if (not hasSOrder):
        idx = [idx[i] if vars[i] != 'sorder' else False for i in range(len(idx))]
    if (not hasTOrder):
        idx = [idx[i] if vars[i] != 'torder' else False for i in range(len(idx))]
    if (not hasOutlier):
        idx = [idx[i] if vars[i] != 'o' and vars[i] != 'ocp' else False for i in range(len(idx))]
    if (not hasData):
        idx = [idx[i] if vars[i] != 'error' else False for i in range(len(idx))]

    vars     = [ vars[i]    for i in range(len(idx)) if idx[i]]
    col      = [ col[i]     for i in range(len(idx)) if idx[i]]
    ylab     = [ ylab[i]    for i in range(len(idx)) if idx[i]]
    heights  = [ heights[i] for i in range(len(idx)) if idx[i]]

    nPlots = length(vars)
    
    if (nPlots == 0):
        error("No valid variable names specified int the 'vars' argument. Possible names include 'st','t','s','sorder','torder','scp','tcp','samp','tslp','o', 'ocp', and 'error'. ")


    #######################################################
    #  Functions and variables to load the outputs
    ########################################################

    opt               = Obj()
    opt.leftmargin    = 0.1;
    opt.rightmargin   = 0.05;
    opt.topmargin     = 0.06;
    opt.bottommargin  = .08;
    opt.verticalspace = 0.01;

    if isempty(fig):
        fig = plt.figure()
    H=axeslayout(opt,heights, fig)

    #  #######################################################
    #  # Create a subplot given the relative heights of the vertical plots
    #  ########################################################
    x      = o
    t, N   = (x.time, length(x.time))
    t2t    = c( [t, flip(t)] )

    for i in range(len(vars)):
        ytitle, var, clr, h = ( ylab[i], vars[i], col[i], H[i])

        #cla(h); % cla
        if var == 'st' :
            Yts, YtsSD, Yerr = get_Yts(x, hasSeason, hasOutlier, hasData)
            plot_st(h, ytitle, has, clr, x, t, t2t, Yts, YtsSD)
        if (var == 's'):
            Y, SD, CI, Amp, AmpSD, Order = get_S(x, hasAmp, hasSOrder);
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_scp(x, ncpStat);
            plot_y(h, ytitle, has, clr, x, t, t2t, Y, CI, ncp, cp);
        if (var == 't'):
            Y, SD, CI, Slp, SlpCI, SlpSignPos, SlpSignZero, Order = get_T(x, hasSlp, hasTOrder);
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_tcp(x, ncpStat);
            plot_y(h, ytitle, has, clr, x, t, t2t, Y, CI, ncp, cp);
        if (var == 'scp'):
            Y, SD, CI, Amp, AmpSD, Order = get_S(x, hasAmp, hasSOrder);
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_scp(x, ncpStat);
            plot_prob(h, ytitle, has, clr, x, t, t2t, Prob1, Prob, ncp, cp);
        if (var == 'tcp'):
            Y, SD, CI, Slp, SlpCI, SlpSignPos, SlpSignZero, Order = get_T(x, hasSlp, hasTOrder);
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_tcp(x, ncpStat);
            plot_prob(h, ytitle, has, clr, x, t, t2t, Prob1, Prob, ncp, cp);
        if (var == 'sorder'):
            Y, SD, CI, Amp, AmpSD, Order = get_S(x, hasAmp, hasSOrder);
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_scp(x, ncpStat);
            plot_order(h, ytitle, has, clr, x, t, t2t, Order, ncp, cp);
        if (var == 'torder'):
            Y, SD, CI, Slp, SlpCI, SlpSignPos, SlpSignZero, Order = get_T(x, hasSlp, hasTOrder);
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_tcp(x, ncpStat);
            plot_order(h, ytitle, has, clr, x, t, t2t, Order, ncp, cp);
        if (var == 'samp'):
            Y, SD, CI, Amp, AmpSD, Order = get_S(x, hasAmp, hasSOrder);
            plot_amp(h, ytitle, has, clr, x, t, t2t, Amp, AmpSD);
        if (var == 'tslp'):
            Y, SD, CI, Slp, SlpCI, SlpSignPos, SlpSignZero, Order = get_T(x, hasSlp, hasTOrder);
            plot_slp(h, ytitle, has, clr, x, t, t2t, Slp, SlpCI)
        if (var == 'slpsgn'):
            Y, SD, CI, Slp, SlpCI, SlpSignPos, SlpSignZero, Order = get_T(x, hasSlp, hasTOrder);
            plot_slpsgn(h, ytitle, has, clr, x, t, t2t, SlpSignPos, SlpSignZero)
        if (var == 'o'):
            Y, SD, CI = get_O(x);
            plot_o(h, ytitle, has, clr, x, t, t2t, Y, SD);
        if (var == 'ocp'):
            cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1 = get_ocp(x, ncpStat);
            plot_oprob(h, ytitle, has, clr, x, t, t2t, Prob1, Prob, ncp, cp);
        if (var == 'error'):
            Yts, YtsSD, Yerr = get_Yts(x, hasSeason, hasOutlier, hasData);
            plot_error(h, ytitle, has, clr, x, t, t2t, Yerr);

        if i % 2 == 0:
            h.yaxis.tick_left()
        else:
            h.yaxis.tick_right()
        #https://stackoverflow.com/questions/63723514/userwarning-fixedformatter-should-only-be-used-together-with-fixedlocator
        #h.set_yticklabels(h.get_yticks(), rotation =90)
        h.tick_params(axis='y', labelrotation = 90)

        if (i == 0):
            fig.suptitle(title);

        if (i == nPlots-1):
            h.set_xlabel(xlabel);
        else:
            h.set_xlabel([]);

        h.set_ylabel(ytitle);
        h.yaxis.set_label_position('left')
        h.yaxis.set_label_coords(-0.08, 0.5)
        #h.yaxis.set_label_position('left')
        fig.show()
    #return {'fig':fig, 'ax':H}
    return fig, H


#                                                                                                                                                    %
def axeslayout(opt, hLayout, fig):
    lm = opt.leftmargin;
    rm = opt.rightmargin;
    tm = opt.topmargin;
    bm = opt.bottommargin;
    vd = opt.verticalspace;
    # hLayout = [1 1 - 1 1 1 - 1 1 1];
    hLayout = np_array(hLayout)
    winList = np_where(hLayout > 0)[0]  # a tuple returned
    hgt = (1 - tm - bm - abs(sum(hLayout[hLayout < 0]) * vd)) / sum(hLayout[hLayout > 0]);
    H = [];
    fig.clf()
    ax = fig.subplots(len(winList), 1)  # plt.figure()  fig.add_axes
    if (len(winList)==1):               # ax is an object not an array
        ax = np_array(ax, dtype=object) 
        ax = ax.reshape(1)              # convert it to an array so ax[0] is valid
    for i in range(len(winList)):
        # H.append( fig.add_axes([0,i/4.0,1,1]) )
        idx = winList[i];
        if i == 0:
            yup = 1 - tm
        else:
            slist = hLayout[arange(1, len(hLayout) + 1) <= idx]
            slist[slist > 0] = slist[slist > 0] * hgt;
            slist[slist < 0] = -slist[slist < 0] * vd;
            yup = 1 - tm - sum(slist);
        #print([lm, yup - hLayout[idx] * hgt, 1 - rm - lm, hLayout[idx] * hgt])

        ax[i].set_position([lm, yup - hLayout[idx] * hgt, 1 - rm - lm, hLayout[idx] * hgt])

        # set(H(i), 'box', 'on');
    return ax

#                                                                                                                                                    %
def get_Yts(x, hasSeason, hasOutlier, hasData):
    Yts = x.trend.Y;
    SD2 = x.trend.SD ** 2 + x.sig2[0]
    if hasSeason:
        Yts, SD2 = (Yts + x.season.Y, SD2 + x.season.SD ** 2)
    if hasOutlier:
        Yts, SD2 = (Yts + x.outlier.Y, SD2 + x.outlier.SD ** 2)
    SD    = np_sqrt(SD2)
    tmp   = Yts + SD
    YtsSD = c([Yts - SD, flip(tmp)])
    if hasData:
        Yerr = x.data - Yts
    else:
        Yerr = [];
    return ((Yts, YtsSD, Yerr))

#                                                                                                                                                    %
def get_T(x, hasSlp, hasTOrder):
    Y   = x.trend.Y;
    SD  = c( [Y - x.trend.SD, flip(Y + x.trend.SD)] )
    if isfield(x.trend, 'CI') and not isempty(x.trend.CI):
        CI  = c( [x.trend.CI[:, 0], flip(x.trend.CI[:,1]) ])
    else:
        CI = SD
    if (hasSlp):
        Slp         = x.trend.slp;
        if isfield(x.trend, 'slpCI') and not isempty(x.trend.slpCI):
            SlpCI    = c( [x.trend.slpCI[:, 0], flip(x.trend.slpCI[:,1]) ])
        else:
            SlpSD = c( [Slp - x.trend.slpSD, flip(Slp + x.trend.slpSD) ] )
            SlpCI = SlpSD 
        SlpSignPos  = x.trend.slpSgnPosPr;
        SlpSignZero = x.trend.slpSgnZeroPr;
    else:
        Slp         = [];
        SlpCI       = [];
        SlpSignPos  = [];
        SlpSignZero = [];

    if (hasTOrder):
        Order = x.trend.order;
    else:
        Order = [];
    return (Y, SD, CI, Slp, SlpCI, SlpSignPos, SlpSignZero, Order)

#                                                                                                                                                    %
def get_tcp(x, ncpStat):
    cmpnt, cp, cpCI = (x.trend, x.trend.cp,x.trend.cpCI)

    if    ncpStat == 'mode':
        ncp = cmpnt.ncp_mode;
    elif  ncpStat == 'median':
        ncp = cmpnt.ncp_median;
    elif  ncpStat == 'mean':
        ncp = cmpnt.ncp;
    elif  ncpStat == 'pct90':
        ncp = cmpnt.ncp_pct90;
    elif ncpStat == 'pct10':
        ncp = cmpnt.ncp_pct10;
    elif ncpStat == 'max':
        ncp = sum(~isnan(cp));
    else:
        ncp = cmpnt.ncp_mode;

    ncp  = int( round(ncp) );
    ncpPr, cpPr,cpChange, Prob = (cmpnt.ncpPr, cmpnt.cpPr ,cmpnt.cpAbruptChange, cmpnt.cpOccPr)
    Prob1 = c( [Prob,Prob - Prob])

    #     ###########################################################
    return (cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1)


def get_S(x, hasAmp, hasSOrder):
    Y = x.season.Y;
    SD = c( [Y - x.season.SD, flip(Y + x.season.SD)])

    if isfield(x.season, 'CI') and not isempty(x.season.CI):
        CI = c( [x.season.CI[:, 0],  flip(x.season.CI[:, 1])] )
    else:
        CI = SD

    if hasAmp:
        Amp   = x.season.amp;
        AmpSD = c( [Amp - x.season.ampSD,  flip(Amp + x.season.ampSD) ])
    else:
        Amp   = [];
        AmpSD = [];

    if hasSOrder:
        Order = x.season.order;
    else:
        Order = [];
    return (Y, SD, CI, Amp, AmpSD, Order)

#     ###########################################################
def get_O(x):
    Y   = x.outlier.Y
    SD = c( [Y - x.outlier.SD,  flip(Y + x.outlier.SD) ])

    if isfield(x.outlier, 'CI') and not isempty(x.outlier.CI):
        CI = c( [x.outlier.CI[:,0], flip(x.outlier.CI[:, 1]) ] )
    else:
        CI = SD
    return (Y, SD, CI)

#                                                                                                                                                    %
def get_scp(x, ncpStat):
    cmpnt, cp, cpCI = (x.season, x.season.cp,x.season.cpCI)

    if    ncpStat == 'mode':
        ncp = cmpnt.ncp_mode;
    elif  ncpStat == 'median':
        ncp = cmpnt.ncp_median;
    elif  ncpStat == 'mean':
        ncp = cmpnt.ncp;
    elif  ncpStat == 'pct90':
        ncp = cmpnt.ncp_pct90;
    elif ncpStat == 'pct10':
        ncp = cmpnt.ncp_pct10;
    elif ncpStat == 'max':
        ncp = sum(~isnan(cp));
    else:
        ncp = cmpnt.ncp_mode;

    ncp  = int( round(ncp) );
    ncpPr, cpPr,cpChange, Prob = (cmpnt.ncpPr, cmpnt.cpPr ,cmpnt.cpAbruptChange, cmpnt.cpOccPr)
    Prob1 = c( [Prob, Prob - Prob])
    #     ###########################################################
    return (cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1)

def  get_ocp(x, ncpStat):
    cmpnt, cp, cpCI = (x.outlier, x.outlier.cp,x.outlier.cpCI)

    if    ncpStat == 'mode':
        ncp = cmpnt.ncp_mode;
    elif  ncpStat == 'median':
        ncp = cmpnt.ncp_median;
    elif  ncpStat == 'mean':
        ncp = cmpnt.ncp;
    elif  ncpStat == 'pct90':
        ncp = cmpnt.ncp_pct90;
    elif ncpStat == 'pct10':
        ncp = cmpnt.ncp_pct10
    elif ncpStat == 'max':
        ncp = sum(~isnan(cp))
    else:
        ncp = cmpnt.ncp_mode

    ncp  = int( round(ncp) );
    ncpPr, cpPr,cpChange, Prob = (cmpnt.ncpPr, cmpnt.cpPr ,[], cmpnt.cpOccPr)
    Prob1 = c( [Prob, Prob - Prob])
    #     ###########################################################
    return (cp, cpCI, ncp, ncpPr, cpPr, cpChange, Prob, Prob1)

#                                                                                                                                                              %
#                                                                                                                                                              %
def getylim(y):
    ymin,ymax = (min(y), max(y) )
    yext = ymax -ymin
    return (ymin - yext*0.1, ymax+yext*0.1)

def plot_st(h, ytitle, has, clr, x, t, t2t, Yts, YtsSD):
    alpha = 0.1;
    h.fill(t2t, YtsSD, alpha=0.3, facecolor=clr.copy(), linestyle='None');
    if has.hasData:
       h.plot(t, x.data, marker='o', color=clr);
    h.plot(t, Yts, color=clr,linestyle='solid',marker='None')
    h.set_ylim( getylim(YtsSD))

#                                                                                                                                                              %
def plot_y(h, ytitle, has, clr, x, t, t2t, Y, CI, ncp, cp):
    alpha = 0.2;
    if (has.hasData  and not has.hasSeason):
        h.plot(t, x.data,  marker='o', color=[.5, .5, .5])
    h.fill(t2t, CI, facecolor=clr, alpha=alpha, linestyle='None');
    h.plot(t, Y, color= clr);
    ylim = h.get_ylim()
    for i in range(ncp):
        h.plot([cp[i], cp[i]], ylim, color='k');
    h.set_ylim(ylim)

#                                                                                                                                                              %
def plot_prob(h, ytitle, has, clr, x, t, t2t, Prob1, Prob, ncp, cp):
    alpha = 0.2;
    h.fill( t2t, Prob1, color=clr, linestyle='None', alpha=alpha);
    h.plot(t, Prob, color=clr)
    maxp = min( [1, max(Prob) * 1.5])
    maxp = max([maxp, 0.2])
    h.set_ylim( (0, maxp) )

    for i in range(ncp):
        h.plot([cp[i], cp[i] ], h.get_ylim(), color='k')

#                                                                                                                                                              %
def plot_order(h, ytitle, has, clr, x, t, t2t, Order, ncp, cp):
    h.plot(t, Order, color=clr);
    maxp = max([ max(Order), 1.05]);
    minp = -0.05;
    h.set_ylim([minp, maxp])
    for i in range(ncp):
        h.plot([cp[i], cp[i]], h.get_ylim(), color='k')

#                                                                                                                                                              %
def plot_amp(h, ytitle, has, clr, x, t, t2t, Amp, AmpSD):
    alpha = 0.5;
    #  $fill(t2t, AmpSD, col=rgb(col[1], col[2], col[3], alpha), border=NA);
    h.plot(t, Amp, color=clr);

#                                                                                                                                                              %
def plot_slp(h, ytitle, has, clr, x, t, t2t, Slp, SlpCI):
    alpha = 0.5;
    h.fill(t2t, SlpCI, color=clr, Linestyle='None', alpha=alpha)
    h.plot(t, Slp, color=clr)

#                                                                                                                                                              %
def plot_slpsgn(h, ytitle, has, clr, x, t, t2t, SlpSignPos, SlpSignZero):
    alpha      = 0.5;
    SlpSignNeg = 1 - SlpSignPos - SlpSignZero;
    y2y        = c( [t - t,  flip( SlpSignNeg ) ] )
    h.fill(t2t, y2y, color=[0, 0, 1], linestyle='None', alpha=alpha);
    y2y = c( [SlpSignNeg,  1 - flip( SlpSignPos) ] )
    h.fill(t2t, y2y, color=[0, 1, 0], linestyle='None', alpha=alpha);
    y2y = c( [1 - SlpSignPos, t - t + 1])
    h.fill(t2t, y2y, color=[1, 0, 0], linestyle='None', alpha=alpha);
    h.plot(t, t - t + 0.5);

def plot_o(h, ytitle, has, clr, x, t, t2t, Y, CI ):
    alpha = 0.5;
    # polygon(t2t, CI,   col  = rgb(col[1],col[2],col[3],alpha), border = NA);
    # points( t,   Y,    type = 'l',col='#333333');
    # TypeError: stem() got an unexpected keyword argument 'use_line_collection'
    # https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.stem.html
    if version < '3.3': # the sign direction changed thanks to KBodolai 
         h.stem( t, Y,  linefmt='-', markerfmt=None ,use_line_collection=True)
    else: 
         h.stem( t, Y,  linefmt='-', markerfmt=None)


def plot_oprob(h, ytitle, has, clr, x, t, t2t, Prob1, Prob, ncp, cp):
    alpha = 0.2
    #% plot(c(t2t[1], t2t), c(0.22, Prob1), type='n', ann=FALSE, xaxt='n', yaxt='n');
   #   # polygon(t2t, Prob1, col  = rgb(col[1],col[2],col[3],alpha), border = NA);
   #    # points( t,   Prob,  col  = rgb(col[1],col[2],col[3])  ,       lwd = 1,type = 'l' );
    if version <  '3.3':  # the sign direction changed thanks to KBodolai 
         h.stem(t, Prob,  linefmt='-', markerfmt=None,use_line_collection=True)
    else: 
         h.stem(t, Prob,  linefmt='-', markerfmt=None)
    

def plot_error(h, ytitle, has, clr, x, t, t2t, Yerr):
    #% plot(t, Yerr, type='n', ann=FALSE, xaxt='n', yaxt='n');
    h.plot(t, t - t, color=clr);
    if version < '3.3':  # the sign direction changed thanks to KBodolai 
         h.stem(t, Yerr, linefmt='-', markerfmt=None,use_line_collection=True)
    else: 
         h.stem(t, Yerr, linefmt='-', markerfmt=None)
    




from .             import Rbeast as cb
from .cvt_to_numpy import force_convert_to_numpy

def beast(Y,                     \
          start            = 1,
          deltat           = 1,
          season           = 'harmonic',  # 'harmonic','dummy','svd','none'
          period           = float('nan'),
          scp_minmax       = [0, 10],     # the min and max numbers of seasonal changepoints
          sorder_minmax    = [0, 5],
          sseg_minlength   = None,    # an integer
          sseg_leftmargin  = None,    # an integer
          sseg_rightmargin = None,    # an integer
          tcp_minmax       = [0, 10], # the min and max numbers of trend changepoints
          torder_minmax    = [0, 1 ],
          tseg_minlength   = None,    # an integer
          tseg_leftmargin  = None,    # an integer
          tseg_rightmargin = None,    # an integer
          method           = 'bayes', # 'bayes','bic','aic','aicc','hic', 'bic0.25','bic0.5'
          detrend          = False,
          deseasonalize    = False,
          mcmc_seed        = 0,
          mcmc_burbin      = 200,
          mcmc_chains      = 3,
          mcmc_thin        = 5,
          mcmc_samples     = 8000,
          precValue        = 1.5,
          precPriorType    = 'componentwise',  # componentwise','uniform','constant','orderwise'
          hasOutlier       = False,
          ocp_minmax       = [0, 10],	  # the min and max numbers of outlier-type changepoints	  
          print_param      = True,
          print_progress   = True,
          print_warning    = True,
          quiet            = False,
          gui              = False,
          dump_ci          = False,
          dump_mcmc        = False,		  
          **kwargs ):
      """
      
      
######################################################################################################
Bayesian changepoint detection and time series decomposition for regular or irregular time series data
    
The fitted model is:
     Y=trend+error             if data has no periodic/seasonal variation (i.e., season='none')
     Y=trend+seasonal+error    if data has periodic/seasonal variation 
     Y=trend+outlier +error    if data is trend-only (no seasonal variation) but with potential outliers
     Y=trend+seasonal+outlier+error if data has periodic/seasonal variation and also has outliers
where trend is a piecewise linear or polynomial function with an unknown number of trend changepoints to 
be inferred; seasonal is a piecewise periodic function with an unknown number of seasonal changepoints to 
be inferred; and the outlier component refers to potential spikes or dips at isolated data points and is 
included only if metadata.hasOutlierCmpnt=True (in beast123) or hasOutlier=True (in beast or beast_irreg)
######################################################################################################
  
--------------------------------------------------------------------------------------------------
*Quick Examples*:
--------------------------------------------------------------------------------------------------
import Rbeast as rb
    
nile,yr =  rb.load_example('nile')             # annual flow of the Nile river
rb.beast( nile, start=1871, season='none' )
    
beach, year = rb.load_example('googletrend')   # the google search trend for 'beach'
o = rb.beast(beach, start= 2004, deltat=1/12, period = 1.0)
rb.plot(o) 
    
--------------------------------------------------------------------------------------------------    
*Input arguments*:
--------------------------------------------------------------------------------------------------
Y:  a regular time series; it should be a numeric vector. For ireggular time series, use 
    'beast_irreg' or 'beast123' instead. For multiple time series or stacked time series 
    images such as satellite data, use 'beast123'.
 
... :  the remaining arguments are many paired keywords and values to specifiy time information 
    or parameters for the beast algorithm. Check the R version of BEAST for detailed explanations
    (https://cran.r-project.org/web/packages/Rbeast/Rbeast.pdf). Below is a brief description.

--------------------------------------------------------------------------------------------------  
*Possible Keywords*:
--------------------------------------------------------------------------------------------------   
start: 
         the start time of the regular time series. Possille values are
         (1) a numeric scalar (e.g., 2002.33); the unit is arbitrary 
         (2) a vector of two values (Year, Month) (e.g., [2002,4] ) 
         (3) a vector of three values (Year, Month, Day) (e.g., [2002,4,15] )
         (4) a date string (e.g., "2002-4-15", "2002/04/15", or "2002/4")
         (5) a datenum as a object variable ( e.g. rb.args( datenum=731320, origin='python'),
             rb.args(year=2002, month=4, day=15) )         
deltat: 
         a number or string; the time interval between consecutive datapoints.. Use a string to specify the time unit
         (e.g., '1/12 year', '1.0 month', '30 days'). Possile units are year,
period:  
         a number or string to specify the period if peridodic/seasonal variations 
         are present in the data. If period is given a zero, negative value or 'none' 
         it suggests no seasonal/periodic component in the signal. (season='none'
         also suggests no periodic component).
         In earlier versions, 'freq' was used to specify the period and
         now deprecated in this version. If period is given as anumber, the unit of 'period', 
         if any, should be consistent with the unit of 'deltat'. If given as a string,
         the unit of period needs to be expicilty specified (e.g., '1 year', '12 mon', '365 days')
season: 
         a string specifier. Possible values - (1) 'none':  trend-only data with no 
         seasonality; (2)'harmonic': the seasonal/peridoic  component modelled via 
         harmonic curves; (3)'dummy': the seasonal component  modelled via a dummy 
         basis (i.e., pulse-like bases); (4)'svd': svd-derived  bases (experimental 
         feature)
scp_minmax: 
         a vector of two integers (e.g.,[0,5]); the min and max number of
         seasonal changepoints allowed
sorder_minmax: 
         a vector of two integers (e.g.,[1,3]); the min and max harmonic orders of
         seasonal changepoints (scp) allowed
sseg_minlength: 
         an integer; the min length of the segment for the seasonal component 
         i.e., the min distance between neighorbing changepoints)
sseg_leftmargin: 
         an integer;  the number of leftmost data points excluded for seasonal changepoint detection.
         That is,  no changepoints are allowed in the starting window/segment of length sseg_leftmargin. 
         sseg_leftmargin must be an unitless integer–the number of time intervals/data points so that the
         time window in the original unit is sseg_leftmargin*deltat. If missing, sseg_leftmargin defaults
         to the minimum segment length 'sseg_min'
sseg_rightmargin: 
         an integer;  the number of rightmost data points excluded for seasonal changepoint detection.
         That is,  no changepoints are allowed in the ending window/segment of length sseg_rightmargin. 
         sseg_rightmargin must be an unitless integer–the number of time intervals/data points so that the
         time window in the original unit is sseg_rightmargin*deltat. If missing, sseg_rightmargin defaults
         to the minimum segment length 'sseg_min'		 
tcp_minmax: 
         a vector of two integers (e.g.,[0,5]); the min and max numbers of
         trend changepoints (tcp) allowed
torder_minmax: 
         a vector of two integers (e.g.,[1,3]); the min and max orders of
         polynomials used to model the trend
tseg_minlength: 
         an integer; the min length of the segment for the trend component (i.e.,
         the min distance between neighorbing changepoints)
tseg_leftmargin: 
         an integer; the number of leftmost data points excluded for trend changepoint detection.
         That is,  no trend changepoints are allowed in the starting window/segment of length tseg_leftmargin. 
         tseg_leftmargin must be an unitless integer–the number of time intervals/data points so that the
         time window in the original unit is tseg_leftmargin*deltat. 
tseg_rightmargin: 
         an integer;  the number of rightmost data points excluded for trend changepoint detection.
         That is,  no trend changepoints are allowed in the ending window/segment of length tseg_rightmargin. 
         tseg_rightmargin must be an unitless integer–the number of time intervals/data points so that the
         time window in the original unit is tseg_rightmargin*deltat.
method: 
         a string specifying which method to formulat model posterior probability. Possible values are
         (1) 'bayes': the full Bayesian formulation (this is the default)  
         (2)'bic':  approximation of posterior probability using the Bayesian information criterion (bic)
         (3)'aic':  approximation of posterior probability using the Akaike information criterion (aic)
         (4)'aicc': approximation of posterior probability using the corrected Akaike information criterion (aicc)
         (5)'hic':  approximation of  posterior probability using the Hannan–Quinn information criterion  (hic)
         (6)'bic0.25':  approximation using the Bayesian information criterion adopted from Kim et al. (2016) <doi: 
              10.1016/j.jspi.2015.09.008>; bic0.25=n*ln(SSE)+0.25k*ln(n) with less complexity penelaty than the standard BIC.
         (7)'bic0.50': the same as above except that the penalty factor is 0.50.
         (8)'bic1.5':  the same as above except that the penalty factor is 1.5.
         (9)'bic2':    the same as above except that the penalty factor is 2.0.
deseasonalize: 
         boolean; if true, the input time series will be first de-seasonalized before applying
         beast by removing a global seasonal component
detrend: 
         boolean; if true, the input time series will be first de-trend before applying 
         beast by removing a global trend  
mcmc_seed: 
         a seed for the random number generator; set it to a non-zero integer to
         reproduce the results among different runs
mcmc_samples: 
         number of MCMC samples collected; the larger, the better
mcmc_thin: 
         a thinning factor for MCMC chains: take every 'mcmc.thin'-th sample
mcmc_burnin: 
         the number of initial samples of each chain to be discarded
mcmc_chains: 
         the number of MCMC chains; the larger, the better but with more computation. 
precValue:
         numeric (>0); the hyperparameter of the precision prior; the default value is 1.5. precValue
         is useful only when precPriorType='constant', as further explained below
precPriorType:
         a string taking one of 'constant', 'uniform',  'componentwise' (default), and 'orderwise'.
         (1) 'constant':  the precision parameter used to parameterize the model coefficients is fixed to
           a const specified by precValue. In other words, precValue is a user-defined hyperparameter 
           and the fitting result may be sensitive to the chosen values of precValue.
         (2) 'uniform':  the precision parameter used to parameterize the model coefficients is a random variable;
           its initial value is specified by precValue. In other words, precValue will be inferred by the MCMC,
           so the fitting result will be insensitive to the chose inital value of precValue.
         (3) 'componentwise': multiple precision parameters are used to parameterize the model coefficients for
           individual components (e.g., one for season and another for trend); their initial values is specified 
           by precValue. In other words, precValue will be inferred by the MCMC, so the fitting result will be 
           insensitive to the choice in precValue.
         (4) 'orderwise'}: multiple precision parameters are used to parameterize the model coefficients not just for 
           individual components but also for individual orders of each component; their initial values is specified 
           by precValue. In other words, precValue will be inferred by the MCMC, so the fitting result will be 
           insensitive to the choice in precValue. 
hasOutlier:
        boolean; if true, the model with an outlier component will be fitted (if season='none',
        Y=trend+outlier+error, or if season ~= 'none', Y=trend+season+outlier+error).		
ocp_minmax:
        a vector of 2 integers (>=0); the min and max numbers of outlier-type changepoints (ocp) allowed in the time series.
        Ocp refers to spikes or dips at isolated times that can't be modeled as trends or seasonal terms.
print_param: 
         boolean; if true, print the beast paramers. 		 
print_progress: 
         boolean; if true, print a progress bar
print_warning: 
         boolean; if true, print warning messages
quiet:
         boolean; if true, supress all the messages and printing		
dump_ci: 
         boolean; if true, credible intervals (i.e., out.season.CI or out.trend.CI) will be computed 
         for the estimated seasonal and trend components. Computing CI is time-consuming, due to sorting, 
         so set dump_ci=Flase if a symmetric credible interval (i.e., out.trend.SD and out.season.SD) suffices.	  		 
dump_mcmc:
         boolean; if true, dump the sampled models in the MCMC chains
gui: 
         boolean; if true, show a gui to demostrate the MCMC sampling; runs only 
         on Windows not Linux or MacOS
		 
######################################################################################################
The keywords for beast() are converted to 'metadata', 'prior','mcmc', and 'extra' options used 
in the beast123() interface. Examples are:
         deseasonalize <-> metadata.deseasonalize
		 hasOutlier    <-> metadata.hasOutlierCmpnt		 
         scp_minmax[0] <-> prior.seasonMinOrder
         scp_minmax[1] <-> prior.seasonMaxOrder
         sseg_min      <-> prior.seasonMinSepDist
         mcmc_seed     <-> mcmc.seed
         tcp_minmax[0] <-> prior.trendMinKnotNumber
         tcp_minmax[1] <-> prior.trendMaxKnotNumber
         dump_ci       <-> extra.computeCredible
Experts should use the the beast123 function.
######################################################################################################		 
 
-------------------------------------------------------------------------------------------------- 
*Result/Output*: The output is a struct variable; example of the fields include
--------------------------------------------------------------------------------------------------
        marg_lik: marginal likilood; the larger, the better
        sig2    : variance  of error
        trend   : the trend component; a struct variable (say, T)
        season  : the season componet; a stuct variable  (say,S)
        The subfields of trend or season:
        .ncpPr        : the prob distribution for number of changepoints
        .ncp          : mean number of changepoints in trend or seasonality
        .ncp_meidan   : median number of changepoints
        .ncp_mode     : mode from ncpPr
        .ncp_pct90    : 90% percentile from ncpPr
        .cpOccPr      : changepoint occurrance probability over time
        .cp           : list of all possible changepoints (many are not sigficant)
        .cpPr         : occurrence probability of the changepoints in cp
        .cpAbruptChange: the sudden changes in trend or seasonlity at cp
        .cpCI         : confidence interval of the cps
        .Y            : the fitted trend or seasonality 
        .SD           : standard deviation of the fitted Y
        .CI           : Credible interval of the fittted Y
        .order   : the mean harmonic or polynomial orders estimated to fit the seasonal and trend     
        trend.slp     : slope of the trend 
        trend.slpSD   : standard dev of the estimated slope
        trend.slpSgnPosPr: time-varying probability of the slope being postive
        trend.slpSgnZeroPr: time-varying probability of the slope being 0
        season.amp     : amplitue of the estiamted seasonality overtime
        season.ampSD   : standard ev of the estiamated amplitude

-------------------------------------------------------------------------------------------------- 
More help:  
--------------------------------------------------------------------------------------------------
       This terse help doc sucks (I know); so far, the best details are still the
       R help doc, available at https://cran.r-project.org/web/packages/Rbeast/Rbeast.pdf.
       Python doesn't allow a '.' in variable names, so Python's equivalent to R's 
       beast(Y,start=1987,tcp.minmax=c(0,5)) is beast(Y, start=1987,  tcp_minmax=[0, 5]).
       
--------------------------------------------------------------------------------------------------
Examples:
--------------------------------------------------------------------------------------------------
import Rbeast as rb

# Nile river annual streamflow: trend-only data
Nile, yr=rb.load_example('nile')      
o = rb.beast(Nile, start=1871, season='none') 
rb.print(o)
rb.plot(o)

# Explicitly specify deltat=1. beast knows nothing about the unit
# of 1871 and 1.0 (i.e., 1871 years, 1871 seconds, or 1871 meters?) 
o=rb.beast(Nile, start=1871, deltat=1.0, season='none') 
 
# start is given a Date 1871-1 (Year-Mon). Then, the time unit is
# fractional/decimal year. delta=1.0 means 1.0 year
o=rb.beast(Nile, start=[1871,1], deltat=1.0, season='none') 
 
# period=0 means a trend-only signal, which is equivalent to season='none'
o=rb.beast(Nile, start=1871, deltat=1, period=0) 
 
# Use a string to specify a unit for deltat or period (e.g., deltat='1 year')
# The time unit is also fractional year. 1871 means Year 1871
o=rb.beast(Nile, start=1871, deltat='1 year', period=0) 
 
# Use a string to specify a unit for delta or period (e.g., deltat='12 mo')
# The time unit is fractional year. 1871 means Year 1871
o=rb.beast(Nile, start=1871, deltat='12 mo', period=0) 
 
#Do not print the options 
o=rb.beast(Nile, start=1871, deltat=1.0, season='none',print_options=False)
 
# Show a gui window to demostrate the beast sampling process in
# real-time (for Windows only not Linux and MacOS)% 
rb.beast(Nile,season='none', gui=True, start=1871) 
 

## Monthly google search popularity of the keyword 'beach'
beach, time = rb.load_example('googletrend')   

# If not specfified, season='harmonic' is assumed by default, with a seasonal 
# component fitted.
o = rb.beast(beach, start=[2004,1], deltat=1/12)  # deltat = 1/12 yr =1 month
rb.print(o)
rb.plot(o)
rb.plot(o,ncpStat='median')
 
# delta  = 1/12: for dates, the default unit is year, so delta=1/12yr=1 month;       
# period = 1.0 means 1 year
o=rb.beast(beach, start='2004-1-15', deltat=1/12, period=1.0)  
 
# period='12 month': use a string to explicitly specify the unit      
o=rb.beast(beach, start=[2004,1], deltat=1/12,      period='12 month')
o=rb.beast(beach, start=[2004,1], deltat='1 month', period='365 days')
 
## Monthly air co2 data since 1959: deltaTime=1/12 year
co2, time = rb.load_example('co2')     
o = rb.beast(co2, start=[1959,1,15],  deltat=1/12,  period=1.0)
rb.print(o)
rb.plot(o)
rb.plot(o,ncpStat='median')
 
## Daily covid-19 infection statistics 
covid=rb.load_example('covid19')    
Y       = covid.newcases
datestr = covid.date
import numpy as np
Y       = np.sqrt(Y)     # a sqrt-root transformation

% the min length of seasonal segments is set to 30 data points
o=rb.beast(Y, start=[2020,1,22],  deltat=1/365,  period=7/365, sseg_min=30)
rb.print(o)
rb.plot(o)
 
# Use a string to specify delta with a unit
o=rb.beast(Y, start=[2020,1,22], deltat='1.0 day',  period='7days', sseg_min=30)
 
# Use a string to specify delta with a unit
o=rb.beast(Y, start=[2020,01,22], deltat='1.0 day',  period='7days', sseg_min=30)
 
Contact info: To report bug or get help, do not hesitate to contact Kaiguang Zhao
at zhao.1423@osu.edu.
      """
    
      Y = force_convert_to_numpy(Y)

    # ......Start of displaying 'MetaData' ......
      metadata = lambda: None  ###Just get an empty object###
                 
    #......Start of displaying 'MetaData' ......
      metadata                  = lambda: None   ###Just get an empty object###
      metadata.isRegularOrdered = True
      metadata.season           = season
      metadata.startTime        = start
      metadata.deltaTime        = deltat
      #metadata.whichDimIsTime   = 1
      metadata.period           = period
      if (season != 'none' and period != period and deltat == deltat):
         if 'freq' in kwargs:
            freq            = kwargs['freq']
            metadata.period = deltat * freq
      metadata.missingValue     = float('nan')
      metadata.maxMissingRate   = 0.7500
      metadata.deseasonalize    = deseasonalize
      metadata.detrend          = detrend
      metadata.hasOutlierCmpnt  = hasOutlier
    #........End of displaying MetaData ........

    #......Start of displaying 'prior' ......
      prior                      = lambda: None   ###Just get an empty object###
      prior.modelPriorType	 = 1
      if season !='none' or season == None:
            prior.seasonMinOrder   = sorder_minmax[0]
            prior.seasonMaxOrder   = sorder_minmax[1]
            prior.seasonMinKnotNum = scp_minmax[0]
            prior.seasonMaxKnotNum = scp_minmax[1]
            prior.seasonMinSepDist = sseg_minlength
            prior.seasonLeftMargin  = sseg_leftmargin
            prior.seasonRightMargin = sseg_rightmargin
      prior.trendMinOrder	  = torder_minmax[0]
      prior.trendMaxOrder	  = torder_minmax[1]
      prior.trendMinKnotNum   = tcp_minmax[0]
      prior.trendMaxKnotNum   = tcp_minmax[1]
      prior.trendMinSepDist   = tseg_minlength
      prior.trendLeftMargin   = tseg_leftmargin
      prior.trendRightMargin  = tseg_rightmargin  
      if hasOutlier:
            prior.outlierMinKnotNum = ocp_minmax[0]	  	  
            prior.outlierMaxKnotNum = ocp_minmax[1]	  
      prior.K_MAX            = 0
      prior.precValue        = precValue
      prior.precPriorType    = precPriorType
    #......End of displaying pripr ......
    #......Start of displaying 'mcmc' ......
      mcmc = lambda: None   ###Just get an empty object###
      mcmc.seed                      = mcmc_seed
      mcmc.samples                   = mcmc_samples
      mcmc.thinningFactor            = mcmc_thin
      mcmc.burnin                    = mcmc_burbin
      mcmc.chainNumber               = mcmc_chains
      mcmc.maxMoveStepSize           = 6
      mcmc.trendResamplingOrderProb  = 0.1000
      mcmc.seasonResamplingOrderProb = 0.1700
      mcmc.credIntervalAlphaLevel    = 0.950
    #......End of displaying mcmc ......
    #......Start of displaying 'extra' ......
      extra = lambda: None   ###Just get an empty object###
      extra.dumpInputData        = True
      extra.whichOutputDimIsTime = 1
      if 'ci' in kwargs:
            dump_ci  = kwargs['ci']      
      extra.computeCredible      = dump_ci
      extra.fastCIComputation    = True
      extra.computeSeasonOrder   = True
      extra.computeTrendOrder    = True
      extra.computeSeasonChngpt  = True
      extra.computeTrendChngpt   = True
      extra.computeSeasonAmp     = False
      extra.computeTrendSlope    = True
      extra.tallyPosNegSeasonJump= False
      extra.tallyPosNegTrendJump = False
      extra.tallyIncDecTrendJump = False
      extra.printProgress        = print_progress
      extra.printParameter       = print_param
      extra.printWarning         = print_warning  
      extra.quiet                = quiet
      extra.dumpMCMCSamples      = dump_mcmc
      extra.consoleWidth         = 85
      extra.numThreadsPerCPU     = 2
      extra.numParThreads        = 0
      if 'cputype' in kwargs.keys():
            extra.cputype = kwargs.get('cputype')
		   
    #......End of displaying extra ......
      if gui:
        o = cb.Rbeast('beastv4demo',Y, metadata, prior, mcmc, extra)
      else:
        #import importlib
        #spec   = importlib.util.spec_from_file_location('Rbeast', 'y:/testold/Rbeast.pyd')
        #module = importlib.util.module_from_spec(spec)
        #spec.loader.exec_module(module)         
        #class xxx:
        #   pass
        #module.setClassObjects(xxx) 
        o = cb.Rbeast('beast_'+method,Y, metadata, prior, mcmc, extra)      
      return (o)




#https://gist.github.com/CMCDragonkai/510ce9456a0429f616baa243d1de3dbf
#https://stackoverflow.com/questions/14295277/not-exporting-functions-from-python-module

#Any directory with __init__.py is considered a package in Python.
#
#Any python files inside a package is considered a module.
#
#Modules contain functions and other bindings that is always exported.
#
#If you are outside the package, and you want to import a module from a package:
#
#from package import module
#module.use_function()

#However this is only because the __init__.py is empty. We can allow the package to export functions and modules as well. Which means this should be possible:

#import package
#package.module.use_function()
#package.use_function()
#To do this, the package's __init__.py must contain something like:

#from . import module
#from .mod import *
#Therefore the __init__.py is kind of like an exporting script for the package. Similar to how I use index.js in my JavaScript #packages. This means you can use the package __init__.py as a sort of staging point for all exports. And anything that isn't# #exported here is hidden. Unless the user really wants to acquire it, in which case they can explicitly use the module.

#To actually have encapsulation at the module level. You need to use _ prefix on your bindings. This is different from __ prefix #used in classes. Note that using the * import will ignore module bindings that have _ prefixed. But they can still be accessed #explicitly if the module is directly accessed. See: https://stackoverflow.com/a/1547160/582917

#I think this form of using __init__.py is the best way. Users shouldn't need to use the from ... import syntax unless they need #to import specific bindings and don't want to use qualified modules.


from .beast        import beast        as beast
from .beast_irreg  import beast_irreg  as beast_irreg
from .beast123     import beast123     as beast123
from .plotbeast    import plot         as plot
from .extractbeast import extract      as extract
from .printbeast   import print        as print, obj_repr
from .load_example import load_example as load_example
from .             import Rbeast       as cbeast
 

# https://stackoverflow.com/questions/2356399/tell-if-python-is-in-interactive-mode
def isInteractiveMode():
    import sys 
    return hasattr(sys, 'ps1') 


# https://stackoverflow.com/questions/67631/how-do-i-import-a-module-given-the-full-path
# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
# https://stackoverflow.com/questions/10675054/how-to-import-a-module-in-python-with-importlib-import-module
# https://stackoverflow.com/questions/63156606/aliases-for-imported-modules-with-importlib
def LoadPlot():
    import sys,importlib, os
    mod       = sys.modules[__name__]    
    mod_path  = os.path.join(os.path.dirname(__file__), 'plotbeast.py')
    spec      = importlib.util.spec_from_file_location("plotbeast", mod_path)
    plotbeast = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(plotbeast)
    #sys.modules["module.name"] = foo    
    setattr(mod,'plot',  getattr(plotbeast, 'plot') )

# https://stackoverflow.com/questions/67631/how-do-i-import-a-module-given-the-full-path
# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
# https://stackoverflow.com/questions/10675054/how-to-import-a-module-in-python-with-importlib-import-module
# https://stackoverflow.com/questions/63156606/aliases-for-imported-modules-with-importlib
def LoadPlot_v2():
    import sys, importlib
    mod       = sys.modules[__name__]    
    plotbeast = importlib.import_module('.plotbeast',__name__)
    setattr(mod,'plot',  getattr(plotbeast, 'plot') ) 

#if isInteractiveMode:
##     LoadPlot()
#      LoadPlot_v2 # this works 


# https://stackoverflow.com/questions/1957780/how-to-override-the-operator-in-python
# https://stackoverflow.com/questions/57413453/is-it-possible-to-override-getitem-at-instance-level-in-python     

class BeastOutputClass:
    pass
BeastOutputClass.__getitem__ = extract    
BeastOutputClass.__repr__    = obj_repr

class args:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self,key, value)
    pass


def RemoveUselessMembers():
    import sys    
    dict = sys.modules[__name__].__dict__.copy()    
    mod  = sys.modules[__name__]
    membersKept = ['args','obj_repr','beast','print','beast_irreg', 'beast123', 'plot','cbeast', '__dict__','BeastOutputClass','load_example']
    for x in dict:
        if x not in membersKept:
            delattr(mod, x)
            
#RemoveUselessMembers()

cbeast.setClassObjects(BeastOutputClass)






from numpy import ndarray, squeeze, array as nparray
from array import array as arrtype

IsTpye = isinstance

def force_convert_to_numpy(x): 
      if  IsTpye(x, list) or IsTpye(x, tuple) or IsTpye(x, arrtype):
           if not IsTpye(x[0], str):    # a sloppy way to test if the array is a string array or not
                 # this is not really needed here bcz they can be converted in the C code
                 return squeeze(nparray(x))
           else:
                 return x   # we assume if the first elm is a sting and the whole array is a string array,
      elif IsTpye(x, ndarray):
           return squeeze( x )
      elif hasattr(x,'to_numpy') and callable(getattr(x,'to_numpy')):
           x = getattr(x,'to_numpy')()
           return squeeze(x)
      else:
          print("WARNING: the data has an uknonwn format but will be forced into numpy.ndarray. If it fails, pls explicilty convert your data into numpy arrays.");
          #raise ValueError('Unknown formats for the input data.')
          return x




import os, sys, glob
from   setuptools import setup, find_packages, Extension,find_namespace_packages

#import numpy as np

# chichen-and-egg problems:  setup should install numpy but numpy is needed to get the numpy headers in np.get_include()
# and run the setup before it is installed
# 
# https://stackoverflow.com/questions/56008251/setup-requires-does-not-seem-to-install-dependencies
# https://stackoverflow.com/questions/35516059/python-is-not-installing-dependencies-listed-in-install-requires-of-setuptools
# https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py/21621689

# A few solutions: https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py/21621689
# This should now (since 2018-ish) be solved by adding numpy as a buildsystem dependency in pyproject.toml, 
# so that pip install makes numpy available before it runs setup.py.

# We fix the problem by using pyproject.toml
# Alternatively, we also included a solution from 
# https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py/21621689
# by using the deferring the import till it has been installed

# https://stackoverflow.com/questions/68282771/python-packaging-build-requirements-in-pyproject-toml-vs-setup-requires
class get_numpy_include(object):
    """Defer numpy.get_include() until after numpy is installed."""
    def __str__(self):
        from numpy import get_include 
        # this is the only function we need here:  not sure if this help reduce 
        # the possibility of "runtimeerror: module compiled against api version 0xe but this version of numpy is 0xd"
        return get_include()


# https://stackoverflow.com/questions/49640513/unicodedecodeerror-charmap-codec-cant-decode-byte-0x9d-in-position-x-charac
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname),encoding='utf8').read()
    #return open(os.path.join(os.path.dirname(__file__), fname)).read()


def is_platform_windows():
    return sys.platform == "win32" or sys.platform == "cygwin"

def is_platform_mac():
    return sys.platform == "darwin"
    

extralibs  = []
if is_platform_windows():
   extralibs=["kernel32", "user32", "gdi32" ]

filenames    = glob.glob('ext_src/*.c');
filenames.append("ext_src/abc_ioFlushcpp.cpp")
      
modules = Extension(
            "Rbeast.Rbeast",
            sources       = filenames,
            #include_dirs  = [ np.get_include(), "ext_src/"],   # Bad option bcz np needs to be installed fist
            #include_dirs  = [get_numpy_include(), "ext_src/"], # Bad option bcz np needs to be installed fist as build-time dependencies                  
            include_dirs  = ["ext_src/"],                 
            define_macros = [('P_RELEASE','1'),('R_INTERFACE','0')],
            libraries     = extralibs
        )
        
#packages = find_packages( include=['exampleproject','exampleproject.*','data'],  exclude=['figures', 'output', 'notebooks'])
#packages = find_packages( exclude=['figures', 'output', 'notebooks'])           
#packages = find_packages( exclude=['tests'])           
packages = find_packages( include =['Rbeast'],exclude=['Rbeast.src'])     
packages = find_namespace_packages(where='./py_src', exclude=['build','tests','extension_src'])
#print(packages)
#print(packages)
setup(
    name             = "Rbeast",   
    version          = '0.1.23',
    description      = "Bayesian changepoint detection and time series decomposition",
    author           = "Kaiguang Zhao",
    author_email     = 'zhao.1423@osu.edu',
    url              = 'https://github.com/zhaokg/Rbeast',    
    keywords         = ['changepoint', 'structural breaks','time series decomposition' 'time series analysis', 'trend analysis'],
    python_requires  = '>=3.5',  
    long_description = read('README.md'),
    long_description_content_type = "text/markdown",        
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",        
        "Programming Language :: Python :: 3.9",                
        "Programming Language :: Python :: 3.10",                        
        "Programming Language :: Python :: 3.11",   
        "Programming Language :: Python :: 3.12",   
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Hydrology",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
    ],    
    #setup_requires      = ['numpy'],           # Deprecated in favor of pyproject.toml
    #install_requires     =  ['numpy>=1.17.3', 'matplotlib>=2.2.0'], # ['numpy>=1.10', 'matplotlib>=2.2.0'],
    install_requires     =  ['numpy>=1.10.0'],                       # remove the depedence on matlplotlib; give a warning if matplot is missing
    #entry_points        ={  'console_scripts': ['mycommand=exampleproject.data:main1'] },    
    packages             = packages,    
    package_dir          = {"Rbeast": "py_src/Rbeast","Rbeast.data": "py_src/Rbeast/data", '': '.'},      
    include_package_data = True,      
    package_data         = {'Rbeast': ['data/nile.csv'],'Rbeast.data': ['googletrend.csv'] ,'Rbeast': ['data/*.npy','data/*.txt','data/*.csv']},
    exclude_package_data = {'': ['*.c','*.cpp']},       
    ext_modules          = [modules]
)



import os, importlib
from numpy import genfromtxt as np_genfromtxt,  load as np_load



def load_example(name):
    """
    Possible example datasets are:
      nile       : annual streamflow of the River Nile from 1871
      googletrend: monthly Google search popularity of the keyword 'beach'
      yellowstone: biweekly NDVI (greenness) of a Yellowstone site
      ohio       : satelite reflectanes and NDVI of an Ohio site
      co2        : monthly CO2 concentrations
      covid19    : weekly  new cases and deaths from COVID-19
      simdata    : a 774x3 matrix: three time series of length 774 
      imagestack : stacked images of NDVI over a small area
    """
    name = name.lower()    
    
    if not (name == 'nile' or name == 'googletrend' or name == 'imagestack' or name == 'ohio'  or 
       name == 'yellowstone' or name == 'co2' or name == 'covid19' or name == 'simdata'):
        print("Possible example datasets are:")    
        print("  nile       : annual streamflow of the River Nile from 1871")    
        print("  googletrend: monthly Google search popularity of the keyword 'beach'")    
        print("  yellowstone: biweekly NDVI (greenness) of a Yellowstone site' ")    
        print("  ohio       : satelite reflectanes and NDVI of an Ohio site  ")    
        print("  co2        : monthly CO2 concentrations  ")    
        print("  covid19    : weekly  new cases and deaths from COVID-19  ")   
        print("  simdata    : a 774x3 matrix: three time series of length 774 ")          
        print("  imagestack : stacked images of NDVI over a small area  \n\n")    
        return None
        
    data_path_csv = os.path.join(os.path.dirname(__file__), 'data', name + '.csv')
    data_path_npy = os.path.join(os.path.dirname(__file__), 'data', name + '.npy')
  
    #https://stackoverflow.com/questions/301134/how-can-i-import-a-module-dynamically-given-its-name-as-string    
    #the code below is not used any longer but kept here as a referenece
    
    # spec = importlib.util.find_spec('pandas')
    # if (spec is None   ):
    #    printf("Error; the 'pandas' package is needed to load the example dataset!\n")            
    # pd = importlib.import_module('pandas')

    try:
        import pandas as pd
        if (name == 'nile'):         
            data = pd.read_csv(data_path_csv)            
            return data.flow, data.year
        if (name == 'googletrend'):         
            data = pd.read_csv(data_path_csv)            
            return data.y, data.time 
        if (name == 'yellowstone'):         
            data = pd.read_csv(data_path_csv)            
            return data.ndvi, data.date
        if (name == 'co2'):         
            data = pd.read_csv(data_path_csv)            
            return data.co2, data.time             
        if (name == 'ohio' or name=='covid19'):         
            data = pd.read_csv(data_path_csv)            
            return data              
    except Exception as error:
        if (name == 'nile' or name == 'googletrend' or name == 'yellowstone' or name == 'co2'):
            x   = np_genfromtxt(data_path_csv, delimiter=',', skip_header=1)
            #data = {'year': x[:, 0], 'flow': x[:, 1]}
            return x[:,1], x[:,0]
            
    if (name == 'simdata'): 
    # np.save('y:/ndvi3d.npy',np.array([a,b,e],dtype='object'),allow_pickle=True)
        x   = np_genfromtxt(data_path_csv, delimiter=',', skip_header=1)
        return x  
        
    if (name == 'imagestack'): 
    # np.save('y:/ndvi3d.npy',np.array([a,b,e],dtype='object'),allow_pickle=True)
        data = np_load(data_path_npy,allow_pickle=True) 
        return data[0], data[1], data[2]   
   
    return None
   
   # the code below is not used any more.
    if (name == 'ndvi'):    
        f1   = os.path.join(os.path.dirname(__file__), 'data',  'ndvi.npy')
        ndvi = np_load(f1)
        f1 = os.path.join(os.path.dirname(__file__), 'data', 'ndvi_fyear.npy')
        year = np_load(f1)
        f1 = os.path.join(os.path.dirname(__file__), 'data', 'ndvi_filename.txt')
        with open(f1) as f:
            datestr = [line.rstrip() for line in f]
        return ndvi, year, datestr

